#!/usr/bin/perl -w
#
#        Name: fci-update-xml-files.pl
#     Version: 0.1
#      Author: Bas Couwenberg <sebastic@xs4all.nl>
#     License: GPL
# Description: Parse the data uploaded by the clients, and the data downloaded 
#              from Stanford, and compile that into serveral XML files.
#              These XML files are used by the web frontend to show the 
#              progress and other information on the clients.
#

use strict;
use Cwd;
use File::Copy;
use Date::Calc;
use Date::Manip;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";

# files:
#  client.xml        per client xml file
#  client-list.xml   list of clients and their per client file
#
#  username.xml      per user xml files
#  username-list.xml list of usernames and their per user file
#
#  team.xml          per team xml file
#  team-list.xml     list of teams and their per team file
#
#  project.xml       per project xml file
#  project-list.xml  list of projects and their per project file

my %cfg;
   $cfg{'self'}                        = 'fci-update-xml-files.pl';
   $cfg{'version'}                     = '0.3';

   $cfg{'www-data-dir'}                = '/var/www/fci/';
   $cfg{'stanford-data-dir'}           = 'stanford-data/';
   $cfg{'client-data-dir'}             = 'client-data/';
   $cfg{'eoc-data-dir'}                = 'eoc-data/';
   $cfg{'xml-data-dir'}                = 'xml-data/';
   
   $cfg{'client-list-file'}            = $cfg{'xml-data-dir'}.'client-list.xml';
   $cfg{'team-list-file'}              = $cfg{'xml-data-dir'}.'team-list.xml';
   $cfg{'project-list-file'}           = $cfg{'xml-data-dir'}.'project-list.xml';
   $cfg{'username-list-file'}          = $cfg{'xml-data-dir'}.'username-list.xml';
   $cfg{'xml-project-summary-file'}    = $cfg{'xml-data-dir'}.'project-summary.xml';
   $cfg{'jmol-projects-file'}          = $cfg{'xml-data-dir'}.'jmol-projects.xml';
   $cfg{'known-projects-file'}         = $cfg{'xml-data-dir'}.'known-projects.xml';
   $cfg{'client-projects-file'}        = $cfg{'xml-data-dir'}.'client-projects.xml';
   $cfg{'fci-summary-file'}            = $cfg{'xml-data-dir'}.'fci-summary.xml';
   $cfg{'eoc-stats-file'}              = $cfg{'xml-data-dir'}.'eoc-stats.xml';
   $cfg{'fahstats-file'}               = $cfg{'xml-data-dir'}.'fahstats.xml';

   $cfg{'user-summary-file'}           = $cfg{'stanford-data-dir'}.'daily_user_summary.txt';
   $cfg{'team-summary-file'}           = $cfg{'stanford-data-dir'}.'daily_team_summary.txt';
   $cfg{'project-summary-file'}        = $cfg{'stanford-data-dir'}.'psummary.html';

   $cfg{verbose}                       = 0,
   $cfg{debug}                         = 0,
   $cfg{quiet}                         = 0,
 
# Get commandline parameters
Getopt::Mixed::init("dir=s
                     verbose
		     debug
		     quiet
		     help
		     d>dir
		     v>verbose
		     b>debug
		     q>quiet
		     h>help
                    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{verbose}        = 1      if($option eq 'verbose');
	$cfg{debug}          = 1      if($option eq 'debug');
	$cfg{quiet}          = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# add directory separator to the end of the www-root path if missing
if($^O eq 'MSWin32') {
        # Windows
        $cfg{'www-data-dir'} .= '\\' if(substr($cfg{'www-data-dir'}, -1, 1) ne '\\');
}
else {
        # One of the UNIX flavors is assumed
        $cfg{'www-data-dir'} .= '/'  if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');
}

# show what's print()ed without waiting for the newline $|=1;
$|=1;

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}
					

# read the XML files  
print "Reading XML files ..." if($cfg{verbose});

my $projects    = XMLin( $cfg{'xml-project-summary-file'},
		         KeyAttr    => 'number',
		         KeepRoot   => 1,
		         ForceArray => [ 'project' ]
		       ) || die $@;

my $jmol_projects = XMLin( $cfg{'jmol-projects-file'},
			   KeyAttr    => { fah_proj => 'number' },
			   ForceArray => [ 'fah_proj' ],
			   KeepRoot   => 1
			 ) || die $@;

my $known_projects = XMLin( $cfg{'known-projects-file'},
			    KeyAttr    => { project => 'number', image => 'type' },
			    ForceArray => [ 'project', 'image' ],
			    KeepRoot   => 1
		          ) || die $@;

my $client_list = XMLin( $cfg{'client-list-file'},
                         KeyAttr    => 'name',
			 KeepRoot   => 1,
			 ForceArray => [ 'client' ]
		       ) || die $@;

my $username_list = XMLin( $cfg{'username-list-file'},
                           KeyAttr    => 'userid',
			   KeepRoot   => 1,
			   ForceArray => [ 'username' ]
			 ) || die $@;

my $team_list   = XMLin( $cfg{'team-list-file'},
                         KeyAttr    => 'number',
		         KeepRoot   => 1,
		         ForceArray => [ 'team', 'username' ]
                       ) || die $@;

print " Done.\n\n" if($cfg{verbose});

my $fci_summary     = {};

my $team            = {};
my $username        = {};
my $client_projects = {};

# get a list of clients
# for every client:
# - process its data
# - saved processed data in per client xml file, client.xml
# - save subset of client data to client-list xml file

my $start_time = get_date();
$fci_summary->{fci_summary}->{last_update}->{start_time} = $start_time;

opendir(DIR, $cfg{'client-data-dir'}) || die "Cannot open directory: $cfg{'client-data-dir'}\n";
foreach my $client_name (readdir DIR) {
	next if($client_name !~ /^[a-zA-Z0-9\_\-]+$/);

	print "Processing client: $client_name\n" if($cfg{verbose});

	my $client_file = $cfg{'client-data-dir'}.$client_name.'/client.xml.new';
	if(!-r $client_file) {
		print "Error: Cannot read new client XML file: $client_file\n" if($cfg{verbose});
		next;
	}

	# client.xml.new content
	my $client = XMLin( $client_file,
	                    ForceArray => [ 'server', 'stats' ]
	                  ) || die $@;

	print "\nChecking last update ...\n" if($cfg{verbose});

	my $client_file_prev = $cfg{'client-data-dir'}.$client_name.'/client.xml';
	if(-r $client_file_prev) {
		my $client_prev = XMLin( $client_file_prev,
					 ForceArray => [ 'server', 'stats' ]
				       );

		my $now      = $start_time;
		my $hour_ago = DateCalc($now, "- 1 hour");
		
		if($cfg{verbose}) {
			print "Previous: $client_prev->{last_update}\n" if($client_prev->{last_update});
			print " Current: $client->{last_update}\n";
			print "     Now: $now\n";
		}

		my $date1 = ParseDate($client->{last_update});
		my $date2 = ParseDate($hour_ago);
		my $flag  = Date_Cmp($date1, $date2);
                        
		my $date3 = ParseDate($now);
		my $delta = DateCalc($date3, $date1, {}, 1);
		   $delta = delta_format($delta, 'long');

		# flag <  0: date1 is earlier than date2
                # flag == 0: date1 and date2 are identical
                # flag >  0: date1 is later than date2
		if($flag < 0) {
			# the client upload is more than 1 hour old
			# consider the client inactive, it has missed it's hourly upload

			print "The last upload is more than 1 hour old.\n" if($cfg{verbose});
			$client->{state} = "inactive";

			my $message = '';
			if($delta !~ /^(\+|\-)/) {
				$message = "Last upload: $delta";
			}

			$client->{state_message} = $message;
		}
		elsif($flag == 0) {
			# the client upload is exacly 1 hour old
			# consider the client active for now, he'll likely be inactive with the next run

			print "The last upload is exactly 1 hour old.\n" if($cfg{verbose});
			$client->{state} = "active";
		}
		else {
			# the client upload less than 1 hour old
			# consider the client active, it has done it's hourly upload

			print "The last upload is less than 1 hour old.\n" if($cfg{verbose});
			$client->{state} = "active";
		}

		if(!$client_prev->{last_update}) {
			# The client has not been processed before (new client)
			
			print "This client has not been processed before.\n" if($cfg{verbose});
			$client->{state} = "new";

			my $message = '';
			if($delta !~ /^(\+|\-)/) {
				$message = "First upload: $delta";
			}

			$client->{state_message} = $message;
		}
	}
	if($cfg{verbose}) {
		print "Client is considered: $client->{state}";
		if(exists $client->{state_message} && $client->{state_message} ne '') {
			print " ($client->{state_message})";
		}
		print "\n\n";
	}
	
	# Replace any pre existing client info from the client-list with the uploaded info,
	# but preserve <first_seen> for clients who have it defined.
	my $first_seen = '';
	if(exists $client_list->{client_list}->{client}->{"$client_name"}) {
		# <first_seen> was introduced in fci-update-xml-files.pl v0.3, if it has not
		# been recorded for clients who were processed with an earlier version, set
		# <first_seen> the clients last update.
		if(defined $client_list->{client_list}->{client}->{"$client_name"}->{first_seen}) {
			$first_seen = $client_list->{client_list}->{client}->{"$client_name"}->{first_seen};
		}
		delete $client_list->{client_list}->{client}->{"$client_name"};
	}
	if($first_seen eq '') {
		$first_seen = $client->{last_update};
	}

	$client_list->{client_list}->{client}->{"$client_name"}->{files}         = $client->{files};
	$client_list->{client_list}->{client}->{"$client_name"}->{last_update}   = $client->{last_update};
	$client_list->{client_list}->{client}->{"$client_name"}->{state}         = $client->{state};
	$client_list->{client_list}->{client}->{"$client_name"}->{state_message} = $client->{state_message};
	$client_list->{client_list}->{client}->{"$client_name"}->{first_seen}    = $first_seen;

	#print Dumper($client_list); 
	#exit;

	my %username; # username.xml content, and username element for username-list.xml

	# add hardware info to client specific xml file
	#if($client_list->{client_list}->{client}->{$client_name}->{hardware}) {
	#	$client->{hardware} = $client_list->{client_list}->{client}->{$client_name}->{hardware};
	#}
	# add Operating System info to client specific xml file
	#if($client_list->{client_list}->{client}->{$client_name}->{operating_system}) {
	#	$client->{operating_system} = $client_list->{client_list}->{client}->{$client_name}->{operating_system};
	#}
	# add fci_version
	#if($client_list->{client_list}->{client}->{$client_name}->{fci_version}) {
	#	$client->{software}->{fci}->{version}
	#		= $client_list->{client_list}->{client}->{$client_name}->{fci_version};
	#}
	# add last_update
	#if($client_list->{client_list}->{client}->{$client_name}->{last_update}) {
	#	$client->{last_update} = $client_list->{client_list}->{client}->{$client_name}->{last_update};
	#}

	# add client.xml to files list
	$client->{files}->{client_xml} = 'client-data/'.$client_name.'/client.xml';
	$client_list->{client_list}->{client}->{"$client_name"}->{files}->{client_xml} = $client->{files}->{client_xml};
	
	if(!-w $client->{files}->{client_xml}) {
		print "Error: Cannot write client XML file! ($client->{files}->{client_xml})\n";
		exit 1;
	}
		
	print "Files for this client:\n" if($cfg{verbose});
	foreach(keys %{$client->{files}}) {
		print " $client->{files}->{$_} ($_)\n" if($cfg{verbose});
		
		# add to the list of files for the client specific xml file
		#$client->{files}->{$_} = $client_list->{client_list}->{client}->{$client_name}->{files}->{$_};
	}
	print "\n" if($cfg{verbose});

	# parse: qd-data.xml.new
	my %qddata;
	if($client->{files}->{qd_data_new}) {
		my $qd_data_file_new = $client->{files}->{qd_data_new};
		
		print "File to parse: $qd_data_file_new\n" if($cfg{verbose});
		if(!-r $qd_data_file_new) {
			print "Error: Cannot read qd-data.xml.new file: $qd_data_file_new\n";
			exit 1;
		}
		%qddata = parse_qddata($qd_data_file_new);
		print "\n" if($cfg{verbose});
	
		# add qd-data.xml to the file list of this client
		$client->{files}->{qd_data} = 'client-data/'.$client_name.'/qd-data.xml';
		$client_list->{client_list}->{client}->{"$client_name"}->{files}->{qd_data} = $client->{files}->{qd_data};
		
		# copy qd-data.xml.new to qd-data.xml 
		copy($qd_data_file_new, $client->{files}->{qd_data}) || die "Error: Cannot copy $qd_data_file_new -> $client->{files}->{qd_data} ($!)\n";
		chmod(0664, $client->{files}->{qd_data});
		
		if(!$qddata{error}) {
			# use the data from qd to update the client and client-list XML files
		
			# add all current work unit info to the client specific xml file
			$client->{current_work_unit} = $qddata{queue}->{index}->{$qddata{queue}->{current_index}};

			# add average stats from qd to current work unit info
			my @stats = qw /download_rate upload_rate performance_fraction pph ppd ppw ppy/;
			foreach(@stats) {
				if($qddata{queue}->{$_}) {
					$client->{stats}->{$_} = $qddata{queue}->{$_};
				}
			}
		
			# only in qd-data.xml:
			# qd release date
			if($qddata{qd}->{date}) {
				# qd info should only go in the client specific file
				$client->{software}->{qd}->{date} = $qddata{qd}->{date};
			}
			# qd functional revision 
			if($qddata{qd}->{functional_revision}) {
				$client->{software}->{qd}->{functional_revision} = $qddata{qd}->{functional_revision};
			}
			# qdinfo.dat release date and maintainer
			if($qddata{qdinfo}) {
				$client->{software}->{qdinfo} = $qddata{qdinfo};
			}
			# current work unit credit 
			# Note: Stanfords psummary and teamstats files call it credit,
			#       whereas qd calls it points.
			#       In qd 046 and before the XML output of qd uses the tags credit and credit_per_hour
			#       In qd 047 and newer  the XML output of qd uses the tags points and points_per_hour
			# FCI uses the name credit in stead of points internally
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit};
				
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit_per_hour}) {
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit_per_hour}
						= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit_per_hour};
				}
			}	
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points};

				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points_per_hour}) {
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit_per_hour}
						= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points_per_hour};
				}

				# $client->{current_work_unit}->{points} should be renamed
				$client->{current_work_unit}->{credit} = $client->{current_work_unit}->{points};
				delete $client->{current_work_unit}->{points};
	
				# $client->{current_work_unit}->{points_per_hour} should be renamed
				if($client->{current_work_unit}->{points_per_hour}) {
					$client->{current_work_unit}->{credit_per_hour} = $client->{current_work_unit}->{points_per_hour};
					delete $client->{current_work_unit}->{points_per_hour};
				}
			}
			# expected time of completion
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{expected}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp};
			}
			# preferred completion deadline
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{preferred}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp};
			}
			# completion deadline
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{deadline}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp};
			}
			# time issued
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{issue_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{issued}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{issue_timestamp};
			}
			# folding info (run, clone, gen, ...)
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{folding}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding};
			}

			# check the progress of the current WU
			#  is it expected before the preferred deadline?
			#  is it expected before the final deadline?
			#
			# States:
			#  ok		expected date before all known deadlines (due, and preferred if known)
			#  warning	expected date after preferred deadline
			#  critical     expected date after final deadline (due)
			#  unknown      expected date is not available, or no deadlines available
			
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp}) {
				my $expect      = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp};
				   $expect      = iso_date($expect);
				my $expect_date = ParseDate($expect);

				my $preferred      = "";
				my $preferred_date = "";
				my $due            = "";
				my $due_date       = "";
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp}) {
					$preferred      = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp};
					$preferred      = iso_date($preferred);
				}
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp}) {
					$due = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp};
					$due = iso_date($due);
				}

				if($preferred ne "" && $due ne "") {
					my $preferred_date = ParseDate($preferred);
					my $due_date       = ParseDate($due);					

					my $preferred_flag = Date_Cmp($expect_date, $preferred_date);
					my $due_flag       = Date_Cmp($expect_date, $due_date);

					# flag <  0: date1 is earlier than date2
					# flag == 0: date1 and date2 are identical
					# flag >  0: date1 is later than date2
					
					if($preferred_flag < 0 && $due_flag < 0) {
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
							= "ok";
						
						$client->{current_work_unit}->{progress_state}
							= "ok";
					}
					elsif($preferred_flag >= 0 && $due_flag < 0) {
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
							= "warning";
						
						$client->{current_work_unit}->{progress_state}
							= "warning";
					}
					elsif($preferred_flag >= 0 && $due_flag >= 0) {
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
							= "critical";
						
						$client->{current_work_unit}->{progress_state}
							= "critical";
					}
					else {
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
							= "unknown";
						
						$client->{current_work_unit}->{progress_state}
							= "unknown";
					}
				}
				elsif($due ne "") {
					my $due_date = ParseDate($due);
					my $due_flag = Date_Cmp($expect_date, $due_date);

					if($due_flag < 0) {
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
							= "ok";
						
						$client->{current_work_unit}->{progress_state}
							= "ok";
					}
					elsif($due_flag >= 0) {
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
							= "critical";
						
						$client->{current_work_unit}->{progress_state}
							= "critical";
					}
					else {
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
							= "unknown";
						
						$client->{current_work_unit}->{progress_state}
							= "unknown";
					}
				}
				else {
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
						= "unknown";
					
					$client->{current_work_unit}->{progress_state}
						= "unknown";
				}
			}
			else {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
					= "unknown";
				
				$client->{current_work_unit}->{progress_state}
					= "unknown";
			}
						
			# also in unitinfo.txt:
			# progress
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{progress}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{progress};
			}
			# project number
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_number}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{project_number}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_number};
			}
			# project name
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_name}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{project_name} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_name};
			}
		
			# also in client.cfg
			# username
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}) {
				$client->{config}->{username}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{username};
				
				$username{name} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{username};
			}
			# team
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{team}) {
				my $team_number = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{team};
				$client->{config}->{team}
					= $team_number;
				
				$team_list->{team_list}->{team}->{$team_number}->{number}
					= $team_number;
				
				$username{team} 
					= $team_number;
			}
			# machineid
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{machine_id}) {
				$client->{config}->{machine_id}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{machine_id};
			}
			# id
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{id}) {
				$client->{config}->{id}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{id};
			}
	
		}
		else {
			print "Warning: Error message in the qd-data for client: $client_name!\n";
			print "Error message: $qddata{error}\n\n";
		}
	}
	else {
		print "Error: No qd-data.xml file for this client!\n";
		print "\nProgram aborted!\n";
		exit 1;
	}

	# parse if available: unitinfo.txt
	my %unitinfo;
	if($client->{files}->{unitinfo}) {
		my $file = $client->{files}->{unitinfo};
	
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			%unitinfo = parse_unitinfo($file);
	
			# project name
			if($unitinfo{project_name} && !$client->{current_work_unit}->{project_name}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{project_name} 
					= $unitinfo{project_name};
				
				$client->{current_work_unit}->{project_name} 
					= $unitinfo{project_name};
			}
			# project_number
			if($unitinfo{project_number} && !$client->{current_work_unit}->{project_number}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{project_number} 
					= $unitinfo{project_number};
					
				$client->{current_work_unit}->{project_number} 
					= $unitinfo{project_number};
			}
			
			# download time
			# not used at the moment
			
			# due time
			# not used at the moment
	
			# progress
			if($unitinfo{progress} && !$client->{current_work_unit}->{progress}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress} 
					= $unitinfo{progress};
				
				$client->{current_work_unit}->{progress} 
					= $unitinfo{progress};
			}
			# progress bar (only in unitinfo.txt)
			if($unitinfo{progress_bar} && length($unitinfo{progress_bar}) == 12) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_bar} 
					= $unitinfo{progress_bar};
				
				$client->{current_work_unit}->{progress_bar} 
					= $unitinfo{progress_bar};
			}
			else {
				print "Warning: Discarding malformed progress bar in unitinfo.txt for client: $client_name\n";
			}
			
			print "\n" if($cfg{verbose});
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{unitinfo};
				delete $client->{files}->{unitinfo};
			}
			
			print "Error: Cannot read unitinfo.txt file: $file\n";
			exit 1;
		}
	}
	
	# parse if available: client.cfg
	my %clientcfg;
	if($client->{files}->{client_cfg}) {
		my $file = $client->{files}->{client_cfg};
		
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			%clientcfg = parse_clientcfg($file);
			print "\n"  if($cfg{verbose});
	
			# The username, team & machineID configured in the client.cfg at the time of
			# WU upload are used to credit the WU, not those recorded in the queue.dat
			# at the time the WU was entered in the queue.

			# username
			if($clientcfg{by_key}{username}) {
				$client->{config}->{username}
					= $clientcfg{by_key}{username};
	
				$username{name} 
					= $clientcfg{by_key}{username};
			}
			# team
			if($clientcfg{by_key}{team}) {
				$client->{config}->{team} 
					= $clientcfg{by_key}{team};

				$team_list->{team_list}->{team}->{$clientcfg{by_key}{team}}->{number}
					= $clientcfg{by_key}{team};
				
				$username{team} 
					= $clientcfg{by_key}{team};
			}
			# machineid
			if($clientcfg{by_key}{machineid}) {
				$client->{config}->{machine_id} 
					= $clientcfg{by_key}{machineid};
			}
			# bigpackets=yes
			if($clientcfg{by_key}{bigpackets}) {
				$client->{config}->{big_wu} 
					= $clientcfg{by_key}{bigpackets};
			}
			else {
				$client->{config}->{big_wu}
					= 'no-pref';
			}
			# type=1
			if($clientcfg{by_key}{type}) {
				$client->{config}->{type}
					= $clientcfg{by_key}{type};
			}
			else {
				$client->{config}->{type}
					= '0';
			}
			# type=1
			if($clientcfg{by_key}{type}) {
				$client->{config}->{type}
					= $clientcfg{by_key}{type};
			}
			else {
				$client->{config}->{type}
					= '0';
			}
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{client_cfg};
				delete $client->{files}->{client_cfg};
			}
			
			print "Error: Cannot read client.cfg file: $file\n";
			exit 1;
		}
	}
	
	# Other files possibly uploaded:
	# FAHlog.txt
	# FAHlog-Prev.txt  /* not used atm */
	# MyFolding.html
	
	# A concatenation of FAHlog.txt and FAHlog-Prev.txt will be saved in
	# FAHlog-Complete.txt
	my $fahlog_complete_file = $cfg{'client-data-dir'}.$client_name.'/FAHlog-Complete.txt';
	
	# FAHlog.txt
	my %fahlog;
	if($client->{files}->{fahlog}) {
		my $file = $client->{files}->{fahlog};
	
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			%fahlog = parse_fahlog($file);
			print "\n" if($cfg{verbose});
			
			#print Dumper(%fahlog);

			$client->{software}->{fah} = $fahlog{client};
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog};
				delete $client->{files}->{fahlog};
			}
			
			print "Error: Cannot read FAHlog.txt file: $file\n";
			exit 1;
		}
	}

	# FAHlog-Prev.txt
	if($client->{files}->{fahlog_prev}) {
		my $file = $client->{files}->{fahlog_prev};
	
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			print "\n" if($cfg{verbose});
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog_prev};
				delete $client->{files}->{fahlog_prev};
			}
		}
	}
	
	# FAHlog-Complete.txt 
	my %fahlog_complete;
	if($client->{files}->{fahlog_complete}) {
		my $file = $client->{files}->{fahlog_complete};

		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			print "Parsing FAHlog-Complete.txt file: $file\n" if($cfg{verbose});
			
			# Get the Time Per Frame info for the Current Work Unit
			if($client->{current_work_unit}->{project_number}                   &&
			   $client->{current_work_unit}->{project_number}        =~ /^\d+$/ &&
			   $client->{current_work_unit}->{folding}                          &&
			   $client->{current_work_unit}->{folding}->{run}        =~ /^\d+$/ &&
			   $client->{current_work_unit}->{folding}->{clone}      =~ /^\d+$/ &&
			   $client->{current_work_unit}->{folding}->{generation} =~ /^\d+$/
			) {
                	        my $project    = $client->{current_work_unit}->{project_number};
                	        my $run        = $client->{current_work_unit}->{folding}->{run};
                	        my $clone      = $client->{current_work_unit}->{folding}->{clone};
                	        my $generation = $client->{current_work_unit}->{folding}->{generation};

				if($cfg{debug}) {
					print "Looking up Time Per Frame lines for:\n";
					print " Project: $project (Run $run, Clone $clone, Gen $generation)\n";
				}
			
				# extract frame times from FAHlog for this WU
				my $i = 0;
		                my @steps;
		                my %seen;
	               		my $total;

		                my $time;
		                my $steps_completed;
		                my $steps_total;
		                my $percentage;
		
		                my $found_project = 0;
				my $prcg          = '';
				
				open(F, $file) || die "Cannot open log file: $file\n";
				while(<F>) {
					chomp;

					s/\r//g; # strip carriage return

		                        # project string to search for:
		                        #
		                        # [13:30:42] Project: 1496 (Run 24, Clone 10, Gen 4)
		                        #
		                        # or for the TINKER core:
		                        #
		                        # [20:51:00] Protein: p1152_L939_K12M_ext_from638
		                        # [20:51:00] - Run: 76 (Clone 46, Gen 2)
		                        # [20:51:00] - Frames Completed: 46, Remaining: 354
		                        # [20:51:00] - Dynamic steps required: 4425000
		                        if(!$found_project) {
		                                if(/^\[(\d{2}:\d{2}:\d{2})\] Project: $project \(Run $run, Clone $clone, Gen $generation\)$/) {
		                                        $found_project = 1;
		                                        $prcg = "$project ($run, $clone, $generation)";
		
		                                        if($cfg{debug}) {
		                                                print "Debug: Found project $prcg\n";
		                                        }
		                                }
		                                elsif($prcg eq $project && /^\[(\d{2}:\d{2}:\d{2})\] - Run: $run \(Clone $clone, Gen $generation\)$/) {
		                                        $found_project = 1;
		                                        $prcg .=  "$project ($run, $clone, $generation)";
		
		                                        if($cfg{debug}) {
		                                                print "Debug: Found project $prcg\n";
		                                        }
		                                }
		                        }
					else {
		                                # Gromacs based cores and Amber/PMD core:
		                                #
		                                # [09:16:00] Completed 311670 out of 500000 steps  (62)
		                                # [10:36:55] Completed 315000 out of 500000 steps  (63)
		                                #
		                                # [09:16:00] Completed 311670 out of 500000 steps  (62%)
		                                # [10:36:55] Completed 315000 out of 500000 steps  (63%)
		                                #
		                                # [09:16:00] Completed 311670 out of 500000 steps  (62 percent)
		                                # [10:36:55] Completed 315000 out of 500000 steps  (63 percent)
		                                #
		                                if(/^\[(\d{2}:\d{2}:\d{2})\] Completed (\d+) out of (\d+) steps  \((\d+).*?\)$/) {
		                                        $time            = $1;
		                                        $steps_completed = $2;
		                                        $steps_total     = $3;
		                                        $percentage      = $4;
		
		                                        if($cfg{debug}) {
		                                                print "Debug: Time: $time; Steps completed: $steps_completed, total: $steps_total; Percentage: $percentage\n";
		                                        }
		                                        # if this is a previously seen step, skip it
		                                        if($seen{$steps_completed}) {
		                                                print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{debug});
		                                                next;
		                                        }
		
		
		                                        push @steps, { time       => $time,
		                                                       completed  => $steps_completed,
		                                                       total      => $steps_total,
		                                                       percentage => $percentage
		                                                     };
		
		                                        # flag this step as seen
		                                        $seen{$steps_completed} = $time;
		
		                                        $i++;
		                                }
		                                # GPU core:
		                                #
		                                # [05:39:50] Completed 10
		                                #
		                                # GPU work/logfile_??.txt can contain:
		                                #
		                                # Completed 10%
		                                #
		                                elsif(/^\[(\d{2}:\d{2}:\d{2})\] Completed (\d+)(\%|)$/) {
		                                        my $s1 = $1;
		                                        my $s2 = $2;
		                                        my $s3 = $3;
		
		                                        $time            = $s1;
		                                        $steps_completed = $s2;
		                                        $steps_total     = 100;
		                                        $percentage      = $s2;
		
		                                        if($cfg{debug}) {
		                                                print "Debug: Time: $time; Steps completed: $steps_completed, total: $steps_total; Percentage: $percentage\n";
		                                        }
		                                        # if this is a previously seen step, skip it
		                                        if($seen{$steps_completed}) {
		                                                print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{debug});
		                                                next;
		                                        }
		
		
		                                        push @steps, { time       => $time,
		                                                       completed  => $steps_completed,
		                                                       total      => $steps_total,
		                                                       percentage => $percentage
		                                                     };
		
		                                        # flag this step as seen
		                                        $seen{$steps_completed} = $time;
		
		                                        $i++;
		                                }
		                                # TINKER core:
		                                #
		                                # [12:37:33] Protein: p1152_L939_K12M_ext_from638
		                                # [12:37:33] - Run: 43 (Clone 14, Gen 22)
		                                # [12:37:33] - Frames Completed: 5, Remaining: 395
		                                # [12:37:33] - Dynamic steps required: 4937500
		                                # ...
		                                # [17:28:23] Finished a frame (12)
		                                # [18:10:00] Finished a frame (13)
		                                #
		                                elsif(/Frames Completed: (\d+), Remaining: (\d+)$/ ||
		                                      /^\[(\d{2}:\d{2}:\d{2})\] Finished a frame \((\d+)\)$/
		                                ) {
		                                        my $s1 = $1;
		                                        my $s2 = $2;
		
		                                        print "Debug: S1: $s1  S2: $s2\n" if($cfg{debug});
		
		                                        # The first regex match is on a sigle digit,
		                                        # so we're parsing a "Frames Completed: ..." line
		                                        if($s1 =~ /^\d+$/) {
		                                                $steps_completed = $s1;
		                                                $steps_total     = $s1 + $s2;
		                                        }
		                                        # the first regex match is a timestamp
		                                        # so we're parsing a "[00:00:00] Finished a frame (0)" line
		                                        elsif($s1 =~ /^\d{2}:\d{2}:\d{2}$/) {
		                                                $time            = $s1;
		                                                $steps_completed = $s2;
		
		                                                # if this is a previously seen step, skip it
		                                                if($seen{$steps_completed}) {
		                                                        print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{debug});
		                                                        next;
		                                                }
		
		                                                push @steps, { time       => $time,
		                                                               completed  => $steps_completed,
		                                                               total      => $steps_total,
		                                                               percentage => $percentage
		                                                             };
		                                                # flag this step as seen
		                                                $seen{$steps_completed} = $time;
			
		                                                $i++;
		
		                                                last if($steps_completed == $steps_total);
		                                        }
						}	
					}
				}
				close F;

				print "Found $i lines with progress information.\n" if($cfg{debug});

		                if($i > 0) {
		                        # calculate the times between the steps
		                        #
		                        # How?: Get the last seen time, and the current time,
		                        #       compare both, while assuming the current time is for
		                        #       today, and the previous time might be yesterday or today
		                        #       (maybe the log spans multiple days).
		                        #       Calculate the difference between those points in time
		                        #       and add that to the list of times.
		
		                        $i = 0;
		
		                        my $previous = 0;
		                        my $current  = 0;
		
		                        my $err;
		                        my $mode = 1;
		                        my $delta;
		
		                        my @times;

		                        foreach(@steps) {
		                                # skip the first time as we can't compare it to his predecessor
		                                if($i == 0) {
		                                        $i++;
		                                        next;
		                                }
		
		                                $previous = $steps[$i-1]->{time};
		                                $current  = $steps[$i]->{time};
		
		                                my $date1 = ParseDate("today ".$previous);
		                                my $date2 = ParseDate("today ".$current);
		                                my $flag  = Date_Cmp($date1, $date2);
		
		                                if($flag < 0) {
		                                        # date1 is earlier, the previous is older than the current
		                                        #
		                                        # the situation may have been:
		                                        #  date1:     today 00:01:02 (previous: 00:01:02)
		                                        #  date2:     today 00:02:03 (current : 00:02:03)
		                                }
		                                elsif($flag == 0) {
		                                        # the two dates are identical
		                                }
		                                else {
		                                        # date2 is earlier, the current is older than the previous
		                                        # the situation may have been:
		                                        #  date1: yesterday 23:59:58 (previous: 23:59:58)
		                                        #  date2:     today 00:01:02 (current : 00:01:02)
		
		                                        $date1 = ParseDate("yesterday ".$previous);

		                                        if($cfg{debug}) {
		                                                print "Warning: The current date in the list is earlier than the previous ($current < $previous)\n";
		                                                print "The previous date was probably yesterday at $previous.\n";
		                                        }

		                                }
		                                print "Parsing dates: $date1 and $date2 ($i).\n" if($cfg{debug});
		
		                                $delta = DateCalc($date1, $date2, \$err, $mode);
		
		                                print "Difference between $date1 and $date2 is $delta.\n" if($cfg{debug});
		
		
		                                # add number of seconds in difference to the list of times per step
		                                my $seconds = -1;
		
		                                #+0:0:0:0:1:55:43
		                                if($delta =~ /(\d+):(\d+):(\d+):(\d+):(\d+):(\d+):(\d+)$/) {
		                                        #+Y:+M:+W:+D:+H:+MN:+S
		                                        my $years   = $1;
		                                        my $months  = $2;
		                                        my $weeks   = $3;
		                                        my $days    = $4;
		                                        my $hours   = $5;
		                                        my $minutes = $6;
		                                           $seconds = $7;
		
		                                        if($cfg{debug}) {
		                                                print "Debug: ";
		                                                print "years: $years ";
		                                                print "months: $months ";
		                                                print "weeks: $weeks ";
		                                                print "days: $days ";
		                                                print "hours: $hours ";
		                                                print "minutes: $minutes ";
		                                                print "seconds: $seconds\n";
		                                        }
							
							#       60   seconds per minute
		                                        $seconds += ($minutes * 60);
							
							#     3600   seconds per hour
		                                        $seconds += ($hours * 60 * 60);

							#    86400   seconds per day
		                                        $seconds += ($days * 24 * 60 * 60);

							#   262800.3 seconds per month
		                                        $seconds += ($months * 30.41667 * 24 * 60 * 60);

							# 31536003
		                                        $seconds += ($years * 12 * 30.41667 * 24 * 60 * 60);
		                                }
		                                print "Debug: Seconds: $seconds\n" if($cfg{debug});
		                                push @times, $seconds;
		
		                                # date1            - date2
		                                # 2006062523:45:28 - 2006062601:41:11 | +0:0:0:0:1:55:43.
		                                #              43 sec +
		                                #  55 min  = 3300 sec +
		                                #   1 hour = 3600 sec =
		                                #            6943 sec
		
		                                $i++;
		                        }
					
					my $min = '';
					my $max = 0;
					my $tot = 0;
					my $avg = 0;
					   $i   = 0;
					
					print "The Times:\n" if($cfg{debug});
					foreach(@times) {
						if($cfg{debug}) {
							print "$i: $_ ";
							print time_format($_);
							print "\n";
						}
						
						if($_ > $max) {
							$max = $_;
						}
						if($min eq '' || ($_ < $min)) {
							$min = $_;
						}
						
						$tot += $_;
						
						$i++;
					}
					$avg = sprintf("%.0f", $tot / $i) if($i > 0);
					
					if($cfg{debug}) {
						print "Max: $max ";
						print time_format($max);
						print "\n";
						print "Min: $min ";
						print time_format($min);
						print "\n";
						print "Tot: $tot ";
						print time_format($tot);
						print "\n";
						print "Avg: $avg ";
						print time_format($avg);
						print "\n";
					}
					
					if(($min ne '' || ($min =~ /^\d+$/ && $min > 0)) &&
					   ($max ne '' || $max > 0) &&
					   ($avg ne '' || $avg > 0) &&
					   ($tot ne '' || $tot > 0) &&
					   ($i > 0)
					) {
						$client->{current_work_unit}->{time_per_frame}->{minimum} = $min;
						$client->{current_work_unit}->{time_per_frame}->{maximum} = $max;
						$client->{current_work_unit}->{time_per_frame}->{average} = $avg;
						$client->{current_work_unit}->{time_per_frame}->{total}   = $tot;
						$client->{current_work_unit}->{time_per_frame}->{frames}  = $i;
	
						$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{time_per_frame}
							= $client->{current_work_unit}->{time_per_frame};
					}
				}
				else {
					print "Warning: No useful log messages found!\n" if($cfg{debug});
				}
			}

			print "\n" if($cfg{verbose});
		}
                else {
                        if(!-e $file) {
                                delete $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog_complete};
                                delete $client->{files}->{fahlog_complete};
                        }

                        print "Error: Cannot read FAHlog-Complete.txt file: $file\n";
                        exit 1;
                }
	}
	
	# MyFolding.html
	my %myfolding;
	if($client->{files}->{myfolding}) {
		my $file = $client->{files}->{myfolding};
	
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			%myfolding = parse_myfolding($file);
			print "\n" if($cfg{verbose});
			
			# username
			if($myfolding{username} && 
			   !$client->{config}->{username}) {
				$client->{config}->{username} 
					= $myfolding{username};
	
				$username{name} 
					= $myfolding{username};
			}
			
			# team
			if($myfolding{team} && 
			   !$client->{config}->{team}) {
				$client->{config}->{team} 
					= $myfolding{team};

				$team_list->{team_list}->{team}->{$myfolding{team}}->{number}
					= $myfolding{team};
	
				$username{team} 
					= $myfolding{team};
			}

			# stats: userpage & teampage
			if($myfolding{statistics}) {
				$username{external_resources} = $myfolding{statistics};

				if($myfolding{statistics}->{teampage} && $client->{config}->{team}) {
					$team_list->{team_list}->{team}->{$client->{config}->{team}}->{external_resources}->{teampage} 
						= $myfolding{statistics}->{teampage};
				}
			}
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{myfolding};
				delete $client->{files}->{myfolding};
			}
			
			print "Error: Cannot read MyFolding.html file: $file\n";
			exit 1;
		}
	}

	# project specific stuff, only usefull when a project number is known
	if($client->{current_work_unit}->{project_number}) {
		my $project_number = $client->{current_work_unit}->{project_number};
		
		# project-summary.xml
		if($projects->{project_summary}->{project}->{$project_number}) {	
			# core code
			if($projects->{project_summary}->{project}->{$project_number}->{code}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{code} 
					= $projects->{project_summary}->{project}->{$project_number}->{code};

				$client->{current_work_unit}->{code}
					= $projects->{project_summary}->{project}->{$project_number}->{code};
			}
		
			# project name
			if($projects->{project_summary}->{project}->{$project_number}->{name}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{name} 
					= $projects->{project_summary}->{project}->{$project_number}->{name};
				
				$client->{current_work_unit}->{project_name}
					= $projects->{project_summary}->{project}->{$project_number}->{name};
			}
		}
	
		# add the current workunit info of this client to the clients per projects list
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{name}
			= $client_name;
		
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{last_update}
			= $client->{last_update};
		
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{state}
			= $client->{state};
		
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{current_work_unit}
			= $client_list->{client_list}->{client}->{"$client_name"}->{current_work_unit};
	}
	
	$client_list->{client_list}->{client}->{"$client_name"}->{config} = $client->{config} if($client->{config});
	$client_list->{client_list}->{client}->{"$client_name"}->{stats}  = $client->{stats}  if($client->{stats});
	
	# add username element to username_list
	if($username{name} && $username{team}) {
		print "Updating Stanford user- & teampage URLs.\n" if($cfg{verbose});
		update_stanford_urls(%username);
		print "\n" if($cfg{verbose});
	}

	#print Dumper($client);
	
	# save the client.xml file
	my $client_xml = $client->{files}->{client_xml};
	
	print "Saving updated client information to: $client_xml\n" if($cfg{verbose});
	XMLout( $client,
	        KeyAttr    => { client => 'name',  server => 'number', fci => 'version' },
	        OutputFile => $client_xml,
	        RootName   => 'client',
		NoAttr     => 1,
	      ) || die $@;

	if($cfg{verbose}) {
		print "\n";
		print "Successfully updated the client XML file for: $client_name\n";
	}

	# update fci summary with number of clients
	$fci_summary->{fci_summary}->{clients} += 1;
	
	print "\n\n" if($cfg{verbose});
} # parsed per client uploaded files


# Add the project data from project-summary.xml to 
# the corresponding client-projects.xml element
print "Adding data from project summary to client projects list.\n" if($cfg{verbose});
foreach my $project (keys %{$client_projects->{client_projects}->{project}}) {
	my $client_list = $client_projects->{client_projects}->{project}->{$project}->{client_list};

	$client_projects->{client_projects}->{project}->{$project} 
		= $projects->{project_summary}->{project}->{$project};

	$client_projects->{client_projects}->{project}->{$project}->{client_list}
		= $client_list;
	
	# update fci summary with number of client projects
	$fci_summary->{fci_summary}->{client_projects} += 1;
}
print "\n" if($cfg{verbose});


# Add the project data from project-summary.xml to 
# the corresponding known-projects.xml element
print "Adding data from project summary to known projects list.\n" if($cfg{verbose});
foreach my $project (keys %{$known_projects->{known_projects}->{project}}) {
	if($projects->{project_summary}->{project}->{$project}) {
		$known_projects->{known_projects}->{project}->{$project}->{project_summary} 
			= $projects->{project_summary}->{project}->{$project};
	}

	if($jmol_projects->{fah_projects}->{fah_proj}->{$project} &&
	   $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{file} &&
	   $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{file} eq 'y'
	) {
		print "Adding URL to Jmol webpage to project: $project\n" if($cfg{verbose});
		$known_projects->{known_projects}->{project}->{$project}->{jmol_url}
			= "http://jmol.sourceforge.net/fah/?project=$project";
	}

	if(exists $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} &&
	   $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} eq 'false'
	) {
		# The project is not currently listed on the psummary page
		# If the project does not have any clients running it, consider it inactive

		if(exists $client_projects->{client_projects}->{project}->{$project} &&
		   exists $client_projects->{client_projects}->{project}->{$project}->{client_list}
		) {
			if(keys %{$client_projects->{client_projects}->{project}->{$project}->{client_list}} == 0) {
				$known_projects->{known_projects}->{project}->{$project}->{state} = 'inactive';
			}
			else {
				$known_projects->{known_projects}->{project}->{$project}->{state} = 'active';
			}
		}
		else {
			$known_projects->{known_projects}->{project}->{$project}->{state} = 'inactive';
		}
	}
	else {
		if(exists $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} &&
		   $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} eq 'true'
		) {
			$known_projects->{known_projects}->{project}->{$project}->{state} = 'active';
		}
		else {
			$known_projects->{known_projects}->{project}->{$project}->{state} = 'inactive';
		}
	}

	# update fci summary with number of client projects
	$fci_summary->{fci_summary}->{known_projects} += 1;
}
print "\n" if($cfg{verbose});


# add the EXTREME Overclocking data to each user
#
# The User Summary URL is in the form:
# http://folding.extremeoverclocking.com/user_summary.php?s=&u=$USERID
#
# The UserID can be found in the users XML file in eoc-data/

print "\nParsing EXTREME Overclocking Folding\@Home Stats data.\n\n" if($cfg{verbose});

foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
	my $username = $username_list->{username_list}->{username}->{$userid}->{name};
	my $team     = $username_list->{username_list}->{username}->{$userid}->{team};

	# Reset the username & team state
	$username_list->{username_list}->{username}->{$userid}->{state} = "unknown";
	$team_list->{team_list}->{team}->{"$team"}->{state} 		= "unknown";

	my $file = $cfg{'eoc-data-dir'}.$username.'-'.$team.'.xml';
	
	print "Looking for XML file for user: $username team: $team ... " if($cfg{verbose});
	
	if(-e $file && -r $file) {
		if($cfg{verbose}) {
			print "Found.\n";
			print "File to parse: $file\n";
		}
		my %eoc_user_summary = parse_eoc_user_summary($file);
		
		# construct the URL to the EOC F@H Stats User Summary page
		if(exists $eoc_user_summary{user}->{UserID} &&
		   $eoc_user_summary{user}->{UserID} =~ /^\d+$/
		) {
			my $url  = "http://folding.extremeoverclocking.com/user_summary.php?s=&u=";
			   $url .= $eoc_user_summary{user}->{UserID};

			$username_list->{username_list}->{username}->{$userid}->{external_resources}->{eoc_user_summary}
				= $url;
		}

		# construct the URL to the EOC F@H Stats Team Summary page
		if(exists $eoc_user_summary{team}->{TeamID} &&
		   $eoc_user_summary{team}->{TeamID} =~ /^\d+$/
		) {
			my $team = $eoc_user_summary{team}->{TeamID};
		
			my $url  = "http://folding.extremeoverclocking.com/team_summary.php?s=&t=";
			   $url .= $team;

			$team_list->{team_list}->{team}->{$team}->{external_resources}->{eoc_team_summary}
				= $url;
		}

		# extract the Points 24 Hour Average for this user
		if(exists $eoc_user_summary{user}->{Points_24hr_Avg} &&
		   $eoc_user_summary{user}->{Points_24hr_Avg} =~ /^\d+(\.\d+)?$/
		) {
			$username_list->{username_list}->{username}->{$userid}->{points_24hr_avg}
				= $eoc_user_summary{user}->{Points_24hr_Avg};

			# Users with a 24 hour average of 0 are considered inactive
			if($eoc_user_summary{user}->{Points_24hr_Avg} == 0) {
				$username_list->{username_list}->{username}->{$userid}->{state}
					= "inactive";
			}
			else {
				$username_list->{username_list}->{username}->{$userid}->{state}
					= "active";
			}
		}

		# extract Points Per Day average for this users team
		if(exists $eoc_user_summary{team}->{Points_24hr_Avg} &&
		   $eoc_user_summary{team}->{Points_24hr_Avg} =~ /^\d+(\.\d+)?$/
		) {
			my $team = $username_list->{username_list}->{username}->{$userid}->{team};
			
			$team_list->{team_list}->{team}->{"$team"}->{points_24hr_avg}
				= $eoc_user_summary{team}->{Points_24hr_Avg};

			# Users with a 24 hour average of 0 are considered inactive
			if($eoc_user_summary{team}->{Points_24hr_Avg} == 0) {
				$team_list->{team_list}->{team}->{"$team"}->{state}
					= "inactive";
			}
			else {
				$team_list->{team_list}->{team}->{"$team"}->{state}
					= "active";
			}
		}
	}
	else {
		if(-e $file && !-r $file) {
			print "Found, but unreadable!\n" if($cfg{verbose});
		}
		else {
			print "Not Found.\n" if($cfg{verbose});
		}
	}
	print "\n" if($cfg{verbose});
}
print "\n" if($cfg{verbose});


# add the kakaostats.com user summary URL to each user
#
# The User Summary URL is in the form:
# http://kakaostats.com/usum.php?u=$USERID
#
# The UserID can be found in the fahstats XML file in xml-data/

if(-r $cfg{'fahstats-file'}) {
	print "Adding kakaostats.com User Summary URLs.\n\n" if($cfg{verbose});

	my $fahstats = XMLin( $cfg{'fahstats-file'},
		              KeyAttr    => { username => 'userid' },
			      KeepRoot   => 1,
			      ForceArray => [ 'username' ]
	                    ) || die "Error: Cannot read fahstats XML file! ($!)";

	foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
		if(exists $fahstats->{fahstats}->{username}->{$userid}->{fahstats_userid} &&
		   $fahstats->{fahstats}->{username}->{$userid}->{fahstats_userid} =~ /^\d+$/
		) {
			my $url  = "http://kakaostats.com/usum.php?u=";
			   $url .= $fahstats->{fahstats}->{username}->{$userid}->{fahstats_userid};
	
			$username_list->{username_list}->{username}->{$userid}->{external_resources}->{fahstats_user_summary}
				= $url;

		}
		
		if(exists $fahstats->{fahstats}->{username}->{$userid}->{team} &&
		   $fahstats->{fahstats}->{username}->{$userid}->{team} =~ /^\d+$/
		) {
			my $team = $fahstats->{fahstats}->{username}->{$userid}->{team};
		
			my $url  = "http://kakaostats.com/tsum.php?t=";
			   $url .= $team;
	
			$team_list->{team_list}->{team}->{$team}->{external_resources}->{fahstats_team_summary}
				= $url;

		}
	}
	print "\n" if($cfg{verbose});
}

# add the Stanford userpage URL to each user
#
# The Stanford userpage URL is in the form:
# http://fah-web.stanford.edu/cgi-bin/main.py?qtype=userpage&username=$USERNAME&teamnum=$TEAMNUMBER
#
# The URL can be found in the MyFolding.html file, 

print "Updating Stanford user- & teampage URLs.\n\n" if($cfg{verbose});
foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
	update_stanford_urls(%{$username_list->{username_list}->{username}->{$userid}});
}
print "\n" if($cfg{verbose});


# update team_list with info from the teamstats file
#   update username_list with info from the teamstats file

my @no_teamstats = ();
foreach my $team (keys %{$team_list->{team_list}->{team}}) {
	my $teamstats_file = $cfg{'xml-data-dir'}.'team-'.$team.'.xml';
	
	print "Looking for stats file for team $team ... " if($cfg{verbose});
	if(-r $teamstats_file) {
		print "Found.\n" if($cfg{verbose});
		
		print "File to parse: $teamstats_file\n" if($cfg{verbose});
		my $xml = XMLin( $teamstats_file,
		                 ForceArray => [ 'rank_entry' ],
				 KeyAttr    => 'team_rank',
				 KeepRoot   => 1,
			       ) || die $@;

		# name
		$team_list->{team_list}->{team}->{$team}->{name} 
			= $xml->{team_stats}->{team}->{name};
		
		# credits
		$team_list->{team_list}->{team}->{$team}->{score}
			= $xml->{team_stats}->{team}->{score};
		
		# work units
		$team_list->{team_list}->{team}->{$team}->{wu}
			= $xml->{team_stats}->{team}->{wu};
		
		# rank
		$team_list->{team_list}->{team}->{$team}->{rank}
			= $xml->{team_stats}->{team}->{rank};
		
		# last_update
		$team_list->{team_list}->{team}->{$team}->{last_update}
			= $xml->{team_stats}->{last_update};

		#print Dumper($xml);
		#print Dumper($username_list);

		my $users = 0;
		foreach my $team_rank (keys %{$xml->{team_stats}->{rank_entry}}) {
			# if the users in the team stats file is one of our fci users
			# update his info
			my $name = $xml->{team_stats}->{rank_entry}->{$team_rank}->{name};

			foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
				my $uname = $username_list->{username_list}->{username}->{$userid}->{name};
				my $uteam = $username_list->{username_list}->{username}->{$userid}->{team};

				if( ($team eq $uteam) && ($name eq $uname) ) {
					print "Parsing entry for user: $name team: $team\n" if($cfg{verbose});
					
					# team_rank, rank inside the team
					$username_list->{username_list}->{username}->{$userid}->{team_rank}
						= $team_rank;

					# rank, rank inside the whole project
					$username_list->{username_list}->{username}->{$userid}->{rank}
						= $xml->{team_stats}->{rank_entry}->{$team_rank}->{rank};

					# credit
					$username_list->{username_list}->{username}->{$userid}->{score}
						= $xml->{team_stats}->{rank_entry}->{$team_rank}->{credit};

					# work units
					$username_list->{username_list}->{username}->{$userid}->{wu}
						= $xml->{team_stats}->{rank_entry}->{$team_rank}->{total};
					
					
					# update fci summary with number of usernames
					$fci_summary->{fci_summary}->{usernames} += 1;
				}
			}
			
			$users++;
		}
		if($users > 0) {
			# members
	 		$team_list->{team_list}->{team}->{$team}->{members} = $users;
		}
	}
	else {
		# in case no teamstats file exists, fall back to daily_*_summary.txt files
		print "Not Found.\n" if($cfg{verbose});
		
		push @no_teamstats, $team;
	}
	
	# update fci summary with number of client projects
	$fci_summary->{fci_summary}->{teams} += 1;
	
	print "\n" if($cfg{verbose});
}


# update team_list with the info from the daily_team_summary.txt file
#   update username_list with the info from the daily_user_summary.txt file
#
# For one or more team, the teamstats where not available
# get the info from daily_team_summary.txt and daily_user_summary.txt

if($#no_teamstats >= 0) {
	print "Looking for daily_team_summary.txt file ... " if($cfg{verbose});
	if(-r $cfg{'team-summary-file'}) {
		print "Found.\n" if($cfg{verbose});
		
		my $i = 0;
		open(F, $cfg{'team-summary-file'}) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			#Tue Aug 30 12:00:00 PDT 2005
			#team    teamname        score   wu
			#446     Google  359611292       4925020
			if(/^(\d+)\t(.*?)\t(\d+)\t(\d+)$/) {
				my $team     = $1;
				my $teamname = $2;
				my $score    = $3;
				my $wu       = $4;

				# The Default and the Google teams both have rank 0
				# as they are special teams, the others have normal ranks
				if($team ne '0' && $team ne '446') {
					$i++;
				}
				
				foreach(@no_teamstats) {
					if($team eq $_) {
						print "Parsing entry for team: $_ name: $teamname\n" if($cfg{verbose});
						
						# name
						$team_list->{team_list}->{team}->{$team}->{name}
							= $teamname;
							
						# credits
						$team_list->{team_list}->{team}->{$team}->{score}
							= $score;
						
						# work units
						$team_list->{team_list}->{team}->{$team}->{wu}
							= $wu;
						
						# rank
						$team_list->{team_list}->{team}->{$team}->{rank}
							= $i;
						
						# members
						# remove the member count since we can only determin that from
						# the teamstats or daily_user_summary.txt
						if($team_list->{team_list}->{team}->{$team}->{members}) {
					 		delete $team_list->{team_list}->{team}->{$team}->{members};
						}
					}
				}
			}
		}
		close F;
	}
	else {
		print "Not Found.\n" if($cfg{verbose});
	}
	print "\n" if($cfg{verbose});

	print "Looking for daily_user_summary.txt file ... " if($cfg{verbose});
	if(-r $cfg{'user-summary-file'}) {
		print "Found.\n" if($cfg{verbose});
		
		open(F, $cfg{'user-summary-file'}) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			#Tue Mar 28 12:00:00 PST 2006
			#name    round(credit)   total   team
			#anonymous       66307146        911359  0
			if(/^(.*?)\t(\d+)\t(\d+)\t(\d+)$/) {
				my $name  = $1;
				my $score = $2;
				my $wu    = $3;
				my $team  = $4;

				foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
					my $uname = $username_list->{username_list}->{username}->{$userid}->{name};
					my $uteam = $username_list->{username_list}->{username}->{$userid}->{team};
					
					my $has_teamstats = 1;
					foreach(@no_teamstats) {
						if($team eq $_) {
							$has_teamstats = 0;
						}
					}

					if( ($team eq $uteam) && ($name eq $uname) && !$has_teamstats ) {
						print "Parsing entry for user: $name team: $team\n" if($cfg{verbose});

						# credit
						$username_list->{username_list}->{username}->{$userid}->{score}
							= $score;
		
						# work units
						$username_list->{username_list}->{username}->{$userid}->{wu}
							= $wu;
						
						
						# update fci summary with number of usernames
						$fci_summary->{fci_summary}->{usernames} += 1;
					}
				}
			}
		}
	}
	else {
		print "Not Found.\n" if($cfg{verbose});
	}
	print "\n" if($cfg{verbose});
}

print "\n" if($cfg{verbose});

#print Dumper($client_list);

# save the client-list.xml file
XMLout( $client_list,
	OutputFile => $cfg{'client-list-file'},
	KeyAttr    => { client => 'name' },
	NoAttr     => 1,
	KeepRoot   => 1
      ) || die $@;

if($cfg{verbose}) {
	print "Successfully updated the client list XML file.\n";
}

# save the team-list.xml file
XMLout( $team_list,
	OutputFile => $cfg{'team-list-file'},
	KeyAttr    => 'number',
	NoAttr     => 1,
	KeepRoot   => 1
      );

if($cfg{verbose}) {
	print "Successfully updated the team list XML file.\n";
}
	
# save the username-list.xml file
XMLout( $username_list,
	OutputFile => $cfg{'username-list-file'},
	KeyAttr    => 'userid',
	NoAttr     => 1,
	KeepRoot   => 1
      ) || die $@;

if($cfg{verbose}) {
	print "Successfully updated the username list XML file.\n";
}

# save the client-projects.xml file
XMLout( $client_projects,
        OutputFile => $cfg{'client-projects-file'},
	KeyAttr    => { project => 'number', client => 'name' },
	NoAttr     => 1,
	KeepRoot   => 1
      ) || die $@;

if($cfg{verbose}) {
	print "Successfully updated the client projects XML file.\n";
}

# save the known-projects.xml file
XMLout( $known_projects,
        OutputFile => $cfg{'known-projects-file'},
	KeyAttr    => { project => 'number', image => 'type' },
	NoAttr     => 1,
	KeepRoot   => 1
      ) || die $@;

if($cfg{verbose}) {
	print "Successfully updated the client projects XML file.\n";
}

# log time of completion in the FCI summary
$fci_summary->{fci_summary}->{last_update}->{end_time} = get_date();

# save the fci-summary.xml file
XMLout( $fci_summary,
        OutputFile => $cfg{'fci-summary-file'},
	NoAttr     => 1,
	KeepRoot   => 1
      ) || die $@;

if($cfg{verbose}) {
	print "Successfully updated the FCI summary XML file.\n";
}
										      

################################[ Subroutines ]################################

##############
#
#     Routine: delta_format()
#  Parameters: Date::Manip::DateCalc($date1, $date2, {}, 1) delta
#     Returns: time in semi human readable format: 1d2h3m4s
# Description: convert delta to readable format
#

sub delta_format {
	my $delta = shift;
	my $type  = shift;

	#+0:0:0:0:1:55:43
        if($delta =~ /^(\+|\-)(\d+):(\d+):(\d+):(\d+):(\d+):(\d+):(\d+)$/) {
                #+Y:+M:+W:+D:+H:+MN:+S
		my $sign    = $1;
                my $years   = $2;
                my $months  = $3;
                my $weeks   = $4;
                my $days    = $5;
                my $hours   = $6;
                my $minutes = $7;
                my $seconds = $8;

                my $time  = '';
		   $time .= $years  ."y"   if($years   > 0);
		   $time .= $months ."M"   if($months  > 0);
		   $time .= $weeks  ."w"   if($weeks   > 0);
		   $time .= $days   ."d"   if($days    > 0);
		   $time .= $hours  ."h"   if($hours   > 0);
		   $time .= $minutes."m"   if($minutes > 0);
		   $time .= $seconds."s"   if($seconds > 0);
		   $time .= ' ago' 	   if($sign eq '-');

		if($type && $type eq 'long') {
			my $long = '';

			if($time =~ /(\d+)y/) {
				my $n = $1;
				$long .= "$n year";
				$long =~ s/year/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)M/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n month";
				$long =~ s/month/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)w/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n week";
				$long =~ s/week/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)d/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n day";
				$long =~ s/day/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)h/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n hour";
				$long =~ s/hour/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)m/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n minute";
				$long =~ s/minute/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)s/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n second";
				$long =~ s/second/$&s/ if($n != 1);
			}
			if($time =~ /ago$/) {
				$long .= " $&";
			}

			$delta = $long;
		}
		else {
			$delta = $time;
		}
	}

	return $delta;
}

##############
#
#     Routine: time_format()
#  Parameters: time in seconds
#     Returns: time in semi human readable format: 1d2h3m4s
# Description: convert time in seconds to readable format
#

sub time_format {
        my $value = shift;

        if($value && ($value >= 0)) {
                # convert time in seconds to human readable format
                # e.g. 65 seconds == 1m5s

                my $date1 = ParseDate("today");
                my $date2 = DateCalc($date1, "+ $value seconds");

                my $delta = DateCalc($date1, $date2, {}, 1);
                #print "Delta: $delta\n";

		$value = delta_format($delta);
        }

        return $value;
}

##############
#
#     Routine: get_date()
#  Parameters: optionally a hash with the keys date and/or time set to something
#     Returns: the date in yyyy-mm-dd format, if only date is set
#              the time in hh:mm:ss format, if only time is set
#              the full date in yyyy-mm-dd hh:mm:ss format, if neither date and
#              time are set
#              both the date and the time, if they're both set
# Description: Get the contents and other info of uploaded data
#

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}

sub parse_qddata {
	my $file = $_[0];
		
	print "Parsing qd-data.xml file: $file\n" if($cfg{verbose});
	
	my %info;
	if(-r $file) {
		my $qd = XMLin( $file,
				KeyAttr    => 'number',
				ForceArray => [ 'index', 'server' ]
			      ) || die $@;

		%info = %{$qd};
	}

	return %info;
}

sub parse_unitinfo {
	my $file = $_[0];

	print "Parsing unitinfo.txt file: $file\n" if($cfg{verbose});

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			# strip windows carriage return character
			if(s/\r//g) {}
			
			#Name: p1134_RIBO_FSpeptide_EXT
			if(/^Name: (.*)$/) {
				$info{project_name} = $1;
				
				if($info{project_name} =~ /^p(\d*)/) {
					$info{project_number} = $1;
				}
			}
			#Download time: January 31 07:33:58
			if(/^Download time: (.*)$/) {
				$info{download} = $1;
			}
			#Due time: February 1 22:44:58
			if(/^Due time: (.*)$/) {
				$info{due} = $1;
			}
			#Progress: 10%  [|_________]
			if(/^Progress: (.*?%)/) {
				$info{progress} = $1;

				if(/^Progress: .*?%\s+(\[.*?\])/) {
					$info{progress_bar} = $1;
				}
			}
		}
		close F;
	}
	
	return %info;
}

sub parse_clientcfg {
	my $file = $_[0];

	print "Parsing client.cfg file: $file\n" if($cfg{verbose});
	
	my %info;
	if(-r $file) {
		my $section = 'undefined';
	
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			# strip windows carriage return character
			if(s/\r//g) {}

			#[settings]
			#username=[DPC]_Fatal_Error_Group0smoking2000
			#team=92
			#asknet=no
			#machineid=1
			#local=217
			#bigpackets=yes
			#passkey=1a2b6789bc2342353653afd2c65u987b
			#
			#[http]
			#active=no
			#host=localhost
			#port=8080
			#usepasswd=yes
			#proxy_name=pu
			#proxy_passwd=Ԯ
			#usereg=yes
			#
			#[power]
			#battery=yes
			#
			#[core]
			#priority=96
			#disableassembly=yes
			#checkpoint=30
			#ignoredeadlines=yes
			#
			#[clienttype]
			#memory=512
			#type=1
			#nonet=yes

			#[settings]
			if(/^\[(.*?)\]$/) {
				$section = $1;
			}
			#username=[DPC]_Fatal_Error_Group0smoking2000
			if(/^(.*?)=(.*)$/) {
				$info{by_section}{$section}{lc($1)} = $2;
				$info{by_key}{lc($1)} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_fahlog {
	my $file = $_[0];

	print "Parsing FAHlog.txt file: $file\n" if($cfg{verbose});
	
	my %info;
	if(-r $file) {
		my $i    = 0;
		my $data = '';
		my @line = ();
	
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			# strip windows carriage return character
			if(s/\r//g) {}
			
			# save the line in the list
			push @line, $_;

			# add the line with line number to variable 
			# to use in regular expressions
			$data .= "$i:".$_."\n";

			# increment the line counter
			$i++;
		}
		
		$i = 0;
		my $current  = '';
		my $previous = '';

                my @steps;
                my %seen;
                my $total;

                my $time;
                my $steps_completed;
                my $steps_total;
                my $percentage;

		foreach(@line) {
			$previous = $current;
			$current  = $_;
			
			#########################
			# Parse log header
			#
			# Example:
			#
			#
			# --- Opening Log file [June 7 10:28:47]
			# 
			# 
			# # Linux Console Edition #######################################################
			# ###############################################################################
			# 
			#                        Folding@Home Client Version 5.02
			# 
			#                           http://folding.stanford.edu
			# 
			# ###############################################################################
			# ###############################################################################
			# 
			# Launch directory: /home/bas/software/folding-schapie
			# Executable: ./FAH502-Linux.exe
			# Arguments: -verbosity 9 -oneunit
			# 
			
			# --- Opening Log file [June 7 10:28:47]
			if(/^.*? Opening Log file \[(.*)\]/) {
				# whenever the client is (re)started it will print this header in the log
				# the client can be restarted many times in a single log 
			
				$info{file_opened} = $1;
			}
			
			# Arguments: -verbosity 9 -configonly
			if(/^Arguments:\s*(.*)\s*/) {
				my $s = $1;
				if($s =~ s/^(.*?)\s*$/$1/){}
				if($s =~ s/-(be|in)\w+//g){}

				#print "arguments : $s\n";
				$info{client}{arguments} = $s;
			}
			
			# Executable: FAH502-Linux.exe
			if(/^Executable:\s*(.*)\s*/) {
				#print "executable : $1\n";
				$info{client}{executable} = $1;
			}
			
			# Launch directory: /home/bas/software/folding
			if(/^Launch directory:\s*(.*)\s*/) {
				#print "directory  : $1\n";
				$info{client}{directory} = $1;
			}
			
			#                        Folding@Home Client Version 5.02
			#                        Folding@Home Client Version 5.90alpha                        
			if(/Folding\@Home Client Version (.*)\s*/) {
				#print "version    : $1\n";
				$info{client}{version} = $1;
			}
			
			# # Windows Console GPU Edition ######
			# # Linux Console Edition ######
			if(/^#\s+(.*)\s+###/) {
				#print "type       : $1\n";
				$info{client}{type} = $1;
			}

		
			#########################
			# Parse client startup & core initialization
			#
			# [19:09:29] - Ask before connecting: No
			# [19:09:29] - User name: [DPC]_Fatal_Error_Group0smoking2000 (Team 92)
			# [19:09:29] - User ID: 722F245B4DE04A25
			# [19:09:29] - Machine ID: 1
			# [19:09:29]
			# [19:09:30] Loaded queue successfully.
			# [19:09:30] + Benchmarking ...
			# [19:09:37] The benchmark result is 437
			# [19:09:37] - Autosending finished units...
			# [19:09:37] Trying to send all finished work units
			# [19:09:37] + No unsent completed units remaining.
			# [19:09:37] - Autosend completed
			# [19:09:37]
			# [19:09:37] + Processing work unit
			# [19:09:37] Core required: FahCore_a0.exe
			# [19:09:37] Core found.
			# [19:09:37] Working on Unit 07 [December 18 19:09:37]
			# [19:09:37] + Working ...
			# [19:09:37] - Calling './FahCore_a0.exe -dir work/ -suffix 07 -checkpoint 15 -verbose -lifeline 25259 -version 502'
			# 

			# [19:09:29] - Ask before connecting: No
			if(/^\[(\d+:\d+:\d+)\] - Ask before connecting: (.*?)$/) {
				$info{config}{ask_before_connecting}{time}  = $1;
				$info{config}{ask_before_connecting}{value} = $2;
			}

			# [16:37:02] - Use IE connection settings: Yes
			if(/^\[(\d+:\d+:\d+)\] - Use IE connection settings: (.*?)$/) {
				$info{config}{use_ie_connection_settings}{time}  = $1;
				$info{config}{use_ie_connection_settings}{value} = $2;
			}
			
			# [19:09:29] - User name: [DPC]_Fatal_Error_Group0smoking2000 (Team 92)
			if(/^\[(\d+:\d+:\d+)\] - User name: (.*?) \(Team (\d+)\)$/) {
				$info{config}{username}{time}  = $1;
				$info{config}{username}{value} = $2;
				
				$info{config}{team}{time}      = $1;
				$info{config}{team}{value}     = $3;
			}
			
			# [19:09:29] - User ID: 722F245B4DE04A25
			if(/^\[(\d+:\d+:\d+)\] - User ID: (.*?)$/) {
				$info{config}{userid}{time}  = $1;
				$info{config}{userid}{value} = $2;
			}
			# [14:15:01] - User ID not found locally
			# [14:15:01] + Requesting User ID from server
			#
			# [14:15:01] - User ID not found locally
			elsif(/^\[(\d+:\d+:\d+)\] - User ID not found locally$/) {
				$info{config}{userid_not_found_locally}{time} = $1;
			}
			# [14:15:01] + Requesting User ID from server
			elsif(/^\[(\d+:\d+:\d+)\] + Requesting User ID from server$/) {
				$info{config}{requesting_userid_from_server}{time} = $1;
			}
			# [10:54:20] - User ID not found locally
			# [10:54:20] + Requesting User ID from server
			# [10:54:20] - Getting ID from AS:
			# [10:54:20] Connecting to http://assign.stanford.edu:8080/
			# [10:54:21] Posted data.
			# [10:54:21] Initial: 7AC9; - Received User ID = 7AC9CAF93C9E3C9B
			#
			# [10:54:20] - Getting ID from AS: 
			elsif(/^\[(\d+:\d+:\d+)\] - Getting ID from AS: (.*)$/) {
				$info{config}{getting_id_from_as}{time}  = $1;
				$info{config}{getting_id_from_as}{value} = $2 ? $2 : "";
			}
			# [10:54:20] Connecting to http://assign.stanford.edu:8080/
			elsif(/^\[(\d+:\d+:\d+)\] Connecting to (.*?)$/) {
				$info{config}{connecting_to}{time}  = $1;
				$info{config}{connecting_to}{value} = $2;
			}
			# [10:54:21] Posted data.
			elsif(/^\[(\d+:\d+:\d+)\] Posted data.$/) {
				$info{config}{posted_data}{time} = $1;
			}
			# [10:54:21] Initial: 7AC9; - Received User ID = 7AC9CAF93C9E3C9B
			elsif(/^\[(\d+:\d+:\d+)\] Initial: (\w+); - Received User ID = (\w+)$/) {
				$info{config}{initial}{time}  = $1;
				$info{config}{initial}{value} = $2;
				
				$info{config}{received_userid}{time}  = $1;
				$info{config}{received_userid}{value} = $3;
			}
			# [14:12:08] - User ID not found locally
			# [14:12:08] + Requesting User ID from server
			# [14:12:36] + Could not connect to Primary Assignment Server for ID
			# [14:13:04] + Could not connect to Secondary Assignment Server for ID
			# [14:13:04] 
			# + Could not get ID from server. Retrying...
			
			# [14:12:36] + Could not connect to Primary Assignment Server for ID
			elsif(/^\[(\d+:\d+:\d+)\] \+ Could not connect to Primary Assignment Server for ID$/) {
				$info{config}{could_not_connect_to_primary_assignment_server_for_id}{time} = $1;
			}
			# [14:13:04] + Could not connect to Secondary Assignment Server for ID
			elsif(/^\[(\d+:\d+:\d+)\] \+ Could not connect to Secondary Assignment Server for ID$/) {
				$info{config}{could_not_connect_to_secondary_assignment_server_for_id}{time} = $1;
			}
			# [14:13:04] 
			elsif(/^\[(\d+:\d+:\d+)\] $/) {
				$info{config}{empty_line}{time} = $1;
			}
			# + Could not get ID from server. Retrying...
			elsif(/^\+ Could not get ID from server. Retrying\.\.\.$/) {
				if($previous =~ /^\[(\d+:\d+:\d+)\] $/) {
					$info{config}{could_not_get_id_from_server_retrying}{time} = $1;
				}
			}

			# [19:09:30] Loaded queue successfully.
			# [19:09:30] + Benchmarking ...
			# [19:09:37] The benchmark result is 437
			
			# [19:09:30] Loaded queue successfully.
			if(/^\[(\d+:\d+:\d+)\] Loaded queue successfully\.$/) {
				$info{client_initialization}{loaded_queue_successfully}{time} = $1;
			}
			# [19:09:30] + Benchmarking ...
			if(/^\[(\d+:\d+:\d+)\] \+ Benchmarking \.\.\.$/) {
				$info{client_initialization}{benchmarking}{time} = $1;
			}
			# [19:09:37] The benchmark result is 437
			if(/^\[(\d+:\d+:\d+)\] The benchmark result is (.*?)$/) {
				$info{client_initialization}{benchmark_result}{time}  = $1;
				$info{client_initialization}{benchmark_result}{value} = $2;
			}
			
			# [19:09:37] - Autosending finished units...
			# [19:09:37] Trying to send all finished work units
			# [19:09:37] + No unsent completed units remaining.
			# [19:09:37] - Autosend completed
			
			# [19:09:37] - Autosending finished units...
			if(/^\[(\d+:\d+:\d+)\] - Autosending finished units\.\.\.$/) {
				$info{client_initialization}{autosending_finished_units}{time} = $1;
			}
			# [19:09:37] Trying to send all finished work units
			if(/^\[(\d+:\d+:\d+)\] Trying to send all finished work units$/) {
				$info{client_initialization}{trying_to_send_all_finished_work_units}{time} = $1;
			}
			# [19:09:37] + No unsent completed units remaining.
			if(/^\[(\d+:\d+:\d+)\] \+ No unsent completed units remaining\.$/) {
				$info{client_initialization}{no_unsent_completed_units_remaining}{time} = $1;
			}
			# [19:09:37] - Autosend completed
			if(/^\[(\d+:\d+:\d+)\] - Autosend completed$/) {
				$info{client_initialization}{autosend_completed}{time} = $1;
			}

			# [19:09:37] + Processing work unit
			# [19:09:37] Core required: FahCore_a0.exe
			# [19:09:37] Core found.
			# [19:09:37] Working on Unit 07 [December 18 19:09:37]
			# [19:09:37] + Working ...
			# [19:09:37] - Calling './FahCore_a0.exe -dir work/ -suffix 07 -checkpoint 15 -verbose -lifeline 25259 -version 502'
			#
			# [18:20:33] + Processing work unit
			# [18:20:33] Core required: FahCore_a0.exe
			# [18:20:33] Core not found.
			# [18:20:33] - Core is not present or corrupted.
			# [18:20:33] - Attempting to download new core...
			 
			# [19:09:37] + Processing work unit
			if(/^\[(\d+:\d+:\d+)\] \+ Processing work unit$/) {
				$info{workunit_initialization}{processing_work_unit}{time} = $1;
			}
			
			# [19:09:37] Core required: FahCore_a0.exe
			if(/^\[(\d+:\d+:\d+)\] Core required: (.*?)$/) {
				$info{workunit_initialization}{core_required}{time}  = $1;
				$info{workunit_initialization}{core_required}{value} = $2;
			}
			
			# [19:09:37] Core found.
			if(/^\[(\d+:\d+:\d+)\] Core found\.$/) {
				$info{workunit_initialization}{core_found}{time} = $1;
			}
			
			# [19:09:37] Working on Unit 07 [December 18 19:09:37]
			if(/^\[(\d+:\d+:\d+)\] Working on Unit (\d+) \[(.*?)\]$/) {
				$info{workunit_initialization}{working_on_unit}{time}  = $1;
				$info{workunit_initialization}{working_on_unit}{value} = $2;
				
				$info{workunit_initialization}{working_on_unit_date}{time}  = $1;
				$info{workunit_initialization}{working_on_unit_date}{value} = $3;
			}
			
			# [19:09:37] + Working ...
			if(/^\[(\d+:\d+:\d+)\] \+ Working \.\.\.$/) {
				$info{workunit_initialization}{working}{time} = $1;
			}
			
			# [19:09:37] - Calling './FahCore_a0.exe -dir work/ -suffix 07 -checkpoint 15 -verbose -lifeline 25259 -version 502'
			if(/^\[(\d+:\d+:\d+)\] - Calling '(.*?)'$/) {
				$info{workunit_initialization}{calling}{time}  = $1;
				$info{workunit_initialization}{calling}{value} = $2;
			}
			
			# [18:20:33] Core not found.
			if(/^\[(\d+:\d+:\d+)\] Core not found\.$/) {
				$info{workunit_initialization}{core_not_found}{time} = $1;
			}
			# [18:20:33] - Core is not present or corrupted.
			elsif(/^\[(\d+:\d+:\d+)\] - Core is not present or corrupted\.$/) {
				$info{workunit_initialization}{core_not_present_or_corrupted}{time} = $1;
			}
			# [18:20:33] - Attempting to download new core...
			elsif(/^\[(\d+:\d+:\d+)\] - Attempting to download new core\.\.\.$/) {
				$info{workunit_initialization}{attempting_to_download_new_core}{time} = $1;
			}

			# [18:20:33] + Downloading new core: FahCore_a0.exe
			# [18:20:33] Downloading core (/~pande/Linux/x86//Core_a0.fah from www.stanford.edu)
			# [18:21:01] Could not download 'www.stanford.edu'
			# [18:21:01] + Error: Could not download core
			# [18:21:01] + Core download error (#2), waiting before retry...
			
			# [18:20:33] + Downloading new core: FahCore_a0.exe
			if(/^\[(\d+:\d+:\d+)\] \+ Downloading new core: (.*?)$/) {
				$info{download_new_core}{downloading_new_core}{time}  = $1;
				$info{download_new_core}{downloading_new_core}{value} = $2;
			}
			# [18:20:33] Downloading core (/~pande/Linux/x86//Core_a0.fah from www.stanford.edu)
			elsif(/^\[(\d+:\d+:\d+)\] Downloading core \((.*?)\)$/) {
				$info{download_new_core}{downloading_core}{time}  = $1;
				$info{download_new_core}{downloading_core}{value} = $2;
			}
			# [18:21:01] Could not download 'www.stanford.edu'
			elsif(/^\[(\d+:\d+:\d+)\] Could not download '(.*?)'$/) {
				$info{download_new_core}{could_not_download}{time}  = $1;
				$info{download_new_core}{could_not_download}{value} = $2;
			}
			# [18:21:01] + Error: Could not download core
			elsif(/^\[(\d+:\d+:\d+)\] \+ Error: Could not download core$/) {
				$info{download_new_core}{error_could_not_download_core}{time} = $1;
			}
			# [18:21:01] + Core download error (#2), waiting before retry...
			elsif(/^\[(\d+:\d+:\d+)\] \+ Core download error \((.*?)\), waiting before retry\.\.\.$/) {
				$info{download_new_core}{core_download_error_waiting_before_retry}{time}   = $1;
				$info{download_new_core}{core_download_error_waiting_before_retry}{value} = $2;
			}
			

			
			#########################
			# Parse frame times
			#
			# Progress formats:
			# 
			# [16:48:21] Completed 70000 out of 125000 steps  (56)
			# [13:02:37] Completed 70000 out of 125000 steps  (56%)
			# [20:39:35] Completed 70000 out of 125000 steps  (56 percent)
			# [19:53:29] Completed 81
			# [19:53:29] Finished a frame (324)
			# [18:35:20] - Frames Completed: 0, Remaining: 400
			#
			#
			# Used formats per core:
			# 
			# Core 0x10: GPU core
			# [19:53:29] Completed 81
			#
			# Core 0x65: Tinker
			# [19:53:29] Finished a frame (324)
			# [18:35:20] - Frames Completed: 0, Remaining: 400
			# 
			# Core 0x78: Gromacs
			# [16:48:21] Completed 70000 out of 125000 steps  (56)
			# [16:05:51] Completed 100000 out of 10000000 steps  (1%)
			# 
			# Core 0x79: Double Gromacs
			# [01:01:34] Completed 100000 out of 125000 steps  (80)
			# [02:38:38] Completed 4900000 out of 5000000 steps  (98%)
			# 
			# Core 0x7A: GB Gromacs
			# [15:34:29] Completed 3350000 out of 5000000 steps  (67)
			# [05:41:45] Completed 14200000 out of 20000000 steps  (71%)
			# 
			# Core 0x80: Gromacs SREM
			# [23:16:44] Completed 165000 out of 1500000 steps  (11%)
			# 
			# Core 0x82: PMD/Amber 
			# [03:56:22] Completed 5000 out of 500000 steps  (1)
			# [04:51:17] Completed 5000 out of 500000 steps  (1%)
			# 
			# Core 0x96: QMD
			# [21:29:20] Completed 1450 out of 2500 steps (58)
			# 
			# Core 0xA0: Gromacs 3.3
			# [03:56:22] Completed 5000 out of 500000 steps  (1)
			# [03:56:22] Completed 5000 out of 500000 steps  (1%)
			# 
			# Core 0xA1: Gromacs SMP
			# [18:17:39] Completed 1650000 out of 5000000 steps  (33 percent)
			#
			# Core 0xCA: Protein Design Core
			# [13:11:45] [SPG] Designing protein sequence 1 of 30
			# [15:49:58] Iterations: 100 of 600
		
			# [16:48:21] Completed 70000 out of 125000 steps  (56)
			# [13:02:37] Completed 70000 out of 125000 steps  (56%)
			# [20:39:35] Completed 70000 out of 125000 steps  (56 percent)
			if(/^\[(\d+:\d+:\d+)\] Completed (\d+) out of (\d+) steps  \((\d+)(|%| percent)\)/) {
				# $time            = $1;
				# $steps_completed = $2;
				# $steps_total     = $3;
				# $percentage      = $4;
				# $indicator       = $5;
			}
			# [19:53:29] Completed 81
			# [19:53:29] Completed 81%
			elsif(/^\[(\d+:\d+:\d+)\] Completed (\d+)(|\%)/) {
				# $time            = $1;
				# $percentage      = $2;
				# $indicator       = $3;
			}
			# [19:53:29] Finished a frame (324)
			elsif(/^\[(\d+:\d+:\d+)\] Finished a frame \((\d+)\)/) {
				# $time            = $1;
				# $frame           = $2;
			}
			# [18:35:20] - Frames Completed: 0, Remaining: 400
			elsif(/^\[(\d+:\d+:\d+)\] - Frames Completed: (\d+), Remaining: (\d+)/) {
				# $time            = $1;
				# $steps_completed = $2;
				# $steps_remaining = $3;
			}
			
			$i++;
		}
		close F;
	}

	return %info;
}

sub parse_myfolding {
	my $file = $_[0];
	
	print "Parsing MyFolding.html file: $file\n" if($cfg{verbose});

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			#User name: [DPC]_Fatal_Error_Group0smoking2000<br>
			if(/^User name: (.*?)\<br\>/) {
				$info{username} = $1;
			}
			
			#Team number: 92<br>
			if(/^ Team number: (\d+)\<br\>/) {
				$info{team} = $1;
			}
			
			#Personal statistics: <a href="http://vspx27.stanford.edu/cgi-bin/main.py?qtype=userpage&username=$USERNAME&teamnum=$TEAM">http://vspx27.stanford.edu/cgi-bin/main.py?qtype=userpage&username=[DPC]_Fatal_Error_Group0smoking2000&teamnum=92</a><br>
			if(/^Personal statistics: \<a href="(.*?)"\>.*?<\/a\>\<br\>/) {
				$info{statistics}->{userpage} = $1;
			}
			
			#Team statistics: <a href="http://vspx27.stanford.edu/cgi-bin/main.py?qtype=teampage&teamnum=92">http://vspx27.stanford.edu/cgi-bin/main.py?qtype=teampage&teamnum=92</a><br>
			if(/^Team statistics: \<a href="(.*?)"\>.*?<\/a\>\<br\>/) {
				$info{statistics}->{teampage} = $1;
			}
		}
		close F;
	}

	return %info;
}

sub parse_eoc_user_summary {
	my $file = $_[0];

	print "Parsing EOC User Summary file: $file\n" if($cfg{verbose});

	my %info;
	if(-r $file) {
		my $xml = XMLin( $file ) || die $@;

		%info = %{$xml};
	}
	
	return %info;
}

sub update_stanford_urls {
	my %username = @_;

	if(!$username{name} || !$username{team}) {
		print "Error: Cannot update Stanford URLs without the username and team!\n";
	}

	# userpage & teampage are only recorded in the MyFolding.html,
	# but the webserver url is hardcoded in the client and outdated on most v5 clients.
	# Overwrite the URLs to point to the correct webserver hostname.
	 
	my $userpage  = "http://fah-web.stanford.edu/cgi-bin/main.py?qtype=userpage&username=";
	   $userpage .= $username{name};
	   $userpage .= "&teamnum=";
	   $userpage .= $username{team};
	
	my $teampage  = "http://fah-web.stanford.edu/cgi-bin/main.py?qtype=teampage&teamnum=";
	   $teampage .= $username{team};
	
	$username{external_resources}->{userpage} = $userpage;
	$username{external_resources}->{teampage} = $teampage;
	
	$team_list->{team_list}->{team}->{$username{team}}->{external_resources}->{teampage} 
		= $teampage;
		
		
	# add username element to username_list
	my $userid = $username{team}.'#'.$username{name};
	$username{userid} = $userid;
		
	$username_list->{username_list}->{username}->{$userid} = \%username;
}

sub iso_date {
        my $date     = $_[0];
        my $iso_date = $date;

        # Tue Jan 17 13:37:47 2006
        if($date =~ /^(\w+)\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\d+)$/) {
                my $dow   = $1;
                my $month = $2;
                my $day   = $3;
                my $time  = $4;
                my $year  = $5;

                $month = Date::Calc::Decode_Month($month);

                $day   = '0'.$day   if(length($day)   == 1);
                $month = '0'.$month if(length($month) == 1);

                $iso_date = "$year-$month-$day $time";
        }

        return $iso_date;
}
	
sub print_usage {
        my %cfg = @_;

        print "$cfg{self} v$cfg{version}\n\n";
        print "Usage: $cfg{self} [OPTIONS]\n\n";
        print "Options:\n";
        print "-d, --dir <DIR>       Path of the FCI server directory\n";
        print "-v, --verbose         Enable verbose output\n";
        print "-q, --quiet           Don't print any output at all, except for errors\n";
        print "-h, --help            Show this usage information\n";
}
